/*
 * Simple demo program for the CI-V interface of the
 * ICOM IC-820H All-Mode Dual Band Satellite radio.
 *
 * Copyright 1996-2000 Dave Dunfield
 * All rights reserved.
 *
 * Compile with DDS MICRO-C/PC, using: cc ic820 -pof
 *
 * Permission granted for personal (non-commercial) use only.
 */
#include <stdio.h>
#include <comm.h>

/*
 * IC-820 CI-V control codes
 */
#define	SYNC		0xFE		// Used to sync to start of message
#define	END			0xFD		// Terminates all transmissions
#define	OK			0xFB		// Command accepted by radio (if no data)
#define	ERROR		0xFA		// Command rejected by radio
#define	READ_BAND	0x02		// Returns LOW to HIGH, same format as FREQ
#define	READ_FREQ	0x03		// Returns LSB to MSB
#define	READ_MODE	0x04		// Returns same as the MODE setting commands
#define	SET_FREQ	0x05		// LSB first, unchanged MSBs can be dropped
#define	MODE_LSB	0x06, 0x00
#define	MODE_USB	0x06, 0x01
#define	MODE_CW		0x06, 0x03, 0x01
#define	MODE_CW_N	0x06, 0x03, 0x02
#define	MODE_FM		0x06, 0x05
#define	VFO			0x07		// VFO mode (last VFO remains selected)
#define	VFO_A		0x07, 0x00	// VFO mode, select VFO A
#define	VFO_B		0x07, 0x01	// VFO mode, select VFO B
#define	VFO_AB		0x07, 0xA0	// Copy active VFO to inactive
#define	MAIN_SUB	0x07, 0xB0	// Swap MAIN and SUB bands
#define	MAIN		0x07, 0xD0	// Control MAIN band
#define	SUB			0x07, 0xD1	// Control SUB band
#define	MEMORY		0x08		// Optional memory channel number can follow
#define	MEMORY_SET	0x09		// Write VFO into memory channel
#define	MEMORY_VFO	0x0A		// Copy memory channel to VFO
#define	MEMORY_CLR	0x0B		// Clear memory channel
#define	READ_OFFSET	0x0C		// Read duplex offset
#define	SET_OFFSET	0x0D		// Follow with offset frequency
#define	SCAN_STOP	0x0E, 0x00
#define	SCAN_START	0x0E, 0x01
#define	SPLIT_OFF	0x0F, 0x00
#define	SPLIT_ON	0x0F, 0x01	// RX on active VFO, TX on inactive
#define	SIMPLEX		0x0F, 0x10
#define	DUPLEX_N	0x0F, 0x11	// Negative offset
#define	DUPLEX_P	0x0F, 0x12	// Postive offset

/*
 * Command table
 */
char *commands[] = {
	"INFORMATION", "PROGRAM", "CODE", "FREQUENCY", "LSB", "USB", "CW",
	"NARROW", "FM", "VFO", "VFA", "VFB", "A=B", "M/S", "MAIN", "SUB",
	"MEMORY", "MW", "M-VFO", "M-CL", "OFFSET", "STOP", "SCAN", "SIMPLEX",
	"-", "+", 0 };

/*
 * Global data
 */
unsigned char
	radio_address = 0x42,		// CI-V address of radio
	control_address = 0xE0,		// CI-V address of controller (us)
	result,						// Result code received from radio
	data[20],					// Data bytes received from radio
	freq[5],					// Frequency input buffer
	com = 1,					// Comm port
	quiet = 0,					// Quiet mode
	debug = 0,					// Debug output
	showres = 0;				// Enables result code display
unsigned
	dcount;						// Count of data bytes received

/*
 * Translation table of mode bytes to ASCII text
 */
static char *mode_text[] = {
	"LSB",						// 0
	"USB",						// 1
	"Unknown(2)",				// 2
	"CW",						// 3 (byte 2 indicates Wide/Narrow)
	"Unknown(4)",				// 4
	"FM",						// 5
	"?" };

/*
 * Issue a multi-byte command to the radio
 */
register radio(int args)
{
	int n, *aptr;

	aptr = (n = nargs()) * 2 + &args;
	start_command();
	while(n--)
		write_chr(*--aptr);
	end_command();
}

/*
 * Issue the start of command header to the radio
 */
start_command()
{
	write_chr(SYNC);
	write_chr(SYNC);
	write_chr(radio_address);
	write_chr(control_address);
}

/*
 * Issue end of command & collect response from radio
 */
end_command()
{
	char *ptr;
	int c;

	ptr = dcount = result = 0;
	write_chr(END);
	if(Cgetct() != SYNC)								// 1st pre-amble
		ptr = "1st Pre-amble not received";
	else if(Cgetct() != SYNC)							// 2nd pre-amble
		ptr = "2nd Pre-amble not received";
	else if(Cgetct() != control_address)				// Controller address
		ptr = "Control address incorrect";
	else if(Cgetct() != radio_address)					// Radio address
		ptr = "Radio address incorrect";
	else if((result = Cgetct()) == ERROR)				// Result code
		ptr = "Radio reported ERROR";
	else while(((c = Cgetct()) != END) && (c != -1))	// Data
		data[dcount++] = c;

	if(ptr) {		// Error was reported... display error & enable results
		printf("***ERROR: %s\n", ptr);
		showres = -1; }

	if(showres) {	// Display result codes
		printf("Result: %02x", result);
		if(dcount) {
			printf(" Data:");
			for(c=0; c < dcount; ++c)
				printf(" %02x", data[c]); }
		printf("\n"); }
}

/*
 * Write a character to the radio, and check the echo
 */
write_chr(int c)
{
	if(debug)
		printf("Write: %02x, ", c);
	Cputc(c);
	if(Cgetct() != c) {
		Cclose();
		abort("Improper echo!"); }
}

/*
 * Get a character from the radio with a timeout
 */
Cgetct()
{
	unsigned t;
	int c;

	t = 0;
	do {
		if((c = Ctestc()) != -1) {
			if(debug)
				printf("Read: %02x\n", c);
			return c; } }
	while ++t;
	return -1;
}

/*
 * Retrieve and display the radio information
 */
show_radio()
{
	/* Interrogate the Main-band */
	radio(MAIN);
	radio(READ_BAND);
	printf("Main band edges are: %x%02x.%02x%02x%02x to %x%02x.%02x%02x%02x\n",
		data[4], data[3], data[2], data[1], data[0],
		data[10], data[9], data[8], data[7], data[6]);
	radio(READ_FREQ);
	printf("Main dial frequency is: %x%02x.%02x%02x%02x\n",
		data[4], data[3], data[2], data[1], data[0]);
	radio(READ_MODE);
	printf("Main mode is: %s%s\n", mode_text[data[0]], (data[1] == 2) ? "-Narrow" : "");

	/* Interrogate the sub-band */
	radio(SUB);
	radio(READ_BAND);
	printf("Sub  band edges are: %x%02x.%02x%02x%02x to %x%02x.%02x%02x%02x\n",
		data[4], data[3], data[2], data[1], data[0],
		data[10], data[9], data[8], data[7], data[6]);
	radio(READ_FREQ);
	printf("Sub  dial frequency is: %x%02x.%02x%02x%02x\n",
		data[4], data[3], data[2], data[1], data[0]);
	radio(READ_MODE);
	printf("Sub  mode is: %s%s\n", mode_text[data[0]], (data[1] == 2) ? "-Narrow" : "");

	/* Back to main band & grab offset */
	radio(MAIN);
	radio(READ_OFFSET);
	printf("Offset setting is: %x.%02x%x.%x\n", data[2], data[1], data[0] >> 4,
		data[0] & 15);
}

/*
 * Parse a frequency from the input line into the freq buffer
 */
char *parse_freq(char *ptr)
{
	int c, digit;

	memset(freq, digit = 0, sizeof(freq));
next: switch(c = *ptr++) {
	case '0' :	case '1' :	case '2' :	case '3' :	case '4' :
	case '5' :	case '6' :	case '7' :	case '8' :	case '9' :
		c -= '0';
		if(!digit) {						/* Mhz (before '.') */
			freq[0] = (freq[0] << 4) + (freq[1] >> 4);
			freq[1] = (freq[1] << 4) + (freq[2] >> 4);
			freq[2] = (freq[2] << 4) + (freq[3] >> 4);
			freq[3] = (freq[3] << 4) + (freq[4] >> 4);
			freq[4] = (freq[4] << 4) + c;
			goto next; }
		if(++digit & 1)						/* Khz (after '.') */
			freq[digit >> 1] = (freq[digit >> 1] & 0xF0) | c;
		else
			freq[digit >> 1] = (freq[digit >> 1] & 0x0F) | (c << 4);
		goto next;
	case '.' :
		freq[0] = freq[3];
		freq[1] = freq[4];
		freq[2] = freq[3] = freq[4] = 0;
		digit = 3;
		goto next; }

	return ptr - 1;
}

/*
 * Program a radio from a file
 *
 * Each line in the file contains one memory channel, which are set
 * sequentially, beginning with channel 1. Each line must contain:
 *	frequency in the form: nnn.nnn
 *	'-' for Negative duplex \_ Neither = Simplex
 *	'+' for Positive duplex /
 *	'F' for FM
 *	'U' for USB
 *	'L' for LSB
 *	'C' for CW
 *	'N' for CW-Narrow
 * Lines beginning with '*' or ';' are ignored (comments).
 * Lines beginning with '$' will prompt with the remainder of the line
 * and then continue (used for functions that must be manualy set).
 */
program_radio(char *filename)
{
	unsigned channel;
	unsigned char buffer[100], *ptr;
	FILE *fp;

	if(!(fp = fopen(filename, "rv")))
		return;

	channel = 0;
	radio(VFO);
	while(fgets(ptr = buffer, sizeof(buffer)-1, fp)) {
		while(isspace(*ptr))
			++ptr;
		switch(*ptr) {
			case '*' :
			case ';' :
			case 0 :
				continue;
			case '$' :
				printf("%s\n..Press ENTER to proceed..\7", buffer+1);
				fgets(buffer, sizeof(buffer)-1, stdin);
				continue; }
		if((++channel & 0x0F) >= 0x0A)				/* Wrap HEX to BCD */
			channel += 6;
		printf("Channel %-2x : ", channel);
		radio(MEMORY, channel);			/* Does not enter MEMORY mode */
		radio(SIMPLEX);
		while(*ptr) switch(*ptr++) {
			case '-' :	radio(DUPLEX_N); 	printf(" -");		continue;
			case '+' :	radio(DUPLEX_P);	printf(" +");		continue;
			case 'F' :	radio(MODE_FM);		printf(" FM");		continue;
			case 'L' :	radio(MODE_LSB);	printf(" LSB");		continue;
			case 'U' :	radio(MODE_USB);	printf(" USB");		continue;
			case 'C' :	radio(MODE_CW);		printf(" CW");		continue;
			case 'N' :	radio(MODE_CW_N);	printf(" CW-N");	continue;
			case '0' :	case '1' :	case '2' :	case '3' :	case '4' :
			case '5' :	case '6' :	case '7' :	case '8' :	case '9' :
				ptr = parse_freq(ptr-1);
				continue;
			case ' ' :	case '\t' :	continue;
			default:
				printf("Invalid character: '%c' in '%s'\n", *(ptr-1), buffer); }
		printf(" %x%02x.%02x%02x%02x\n",
			freq[0], freq[1], freq[2], freq[3], freq[4]);
		radio(SET_FREQ, freq[4], freq[3], freq[2], freq[1], freq[0]);
 		radio(MEMORY_SET); }

	fclose(fp);
}

/*
 * Main program
 */
main(int argc, char *argv[])
{
	int i, c, arg;
	char *ptr;

	/* Process command line arguments */
	arg = 1;
getopt:
	if((arg < argc) && (*(ptr = argv[arg]) == '/')) {
		++arg;
		while(*++ptr) switch(toupper(*ptr)) {
			case '1' :	com = 1;		continue;
			case '2' :	com = 2;		continue;
			case '3' :	com = 3;		continue;
			case '4' :	com = 4;		continue;
			case 'D' :	debug = -1;		continue;
			case 'Q' :	quiet = -1;		continue;
			case 'R' :	showres = -1;	continue;
			default:
				printf("\nInvalid option: %c\n", *ptr);
				return; }
		goto getopt; }

	if(!quiet)
		printf("\nIC820 CI-V DEMO program.  Copyright 1996-2000 Dave Dunfield\n");

	/* If no arguments, show command summary */
	if(arg >= argc) {
		printf("\nUse: IC820 /1-4 /D /Q /R command [options]\n\nAvailable commands:\n\n");
		for(i=0; ptr=commands[i];) {
			if(++i % 6)
				printf("%-12s", ptr);
			else
				printf("%s\n", ptr); }
		return; }

	/* Search for command in command list */
	strupr(argv[arg]);				// convert to upper case for search
	i = c = 0;
	while(ptr = commands[i++])
		if(strbeg(ptr, argv[arg])) {	// We have found it
			if(c)					// TWICE! .. ambiguous
				abort("\nAmbiguous command!\n");
			c = i; }

	if(!c) {
		printf("\nInvalid command: %s\n", argv[arg]);
		return; }
	++arg;

	if(!quiet)
		printf("Command: %s\n", commands[c-1]);

	/* Open the comm port at ICOM standard 1200 BPS */
	if(Copen(com, _1200, PAR_NO|DATA_8|STOP_1, SET_RTS|SET_DTR|OUTPUT_2))
		abort("Cannot open comm port");
	Cflags |= TRANSPARENT;
	delay(50);

	/* Process command */
	switch(c) {
	case 1 :	// Information
		show_radio();
		break;
	case 2 :	// Program
		if(arg >= argc)
			printf("\nFilename required!\n");
		else
			program_radio(argv[arg]);
		break;
	case 3 :	// Code
		if(arg >= argc)
			printf("\nCI-V code(s) required!\n");
		else {
			start_command();
			while(arg < argc) {
				c = atox(argv[arg++]);
				write_chr(c); }
			end_command(); }
		break;
	case 4 :	// Frequency
		if(arg >= argc)
			printf("\nFrequency required!\n");
		else {
			parse_freq(argv[arg]);
			radio(SET_FREQ, freq[4], freq[3], freq[2], freq[1], freq[0]); }
		break;
	case 5 :	radio(MODE_LSB);		break;
	case 6 :	radio(MODE_USB);		break;
	case 7 :	radio(MODE_CW);			break;
	case 8 :	radio(MODE_CW_N);		break;
	case 9 :	radio(MODE_FM);			break;
	case 10 :	radio(VFO);				break;
	case 11 :	radio(VFO_A);			break;
	case 12 :	radio(VFO_B);			break;
	case 13 :	radio(VFO_AB);			break;
	case 14 :	radio(MAIN_SUB);		break;
	case 15 :	radio(MAIN);			break;
	case 16 :	radio(SUB);				break;
	case 17 :	// Memory
		if(arg >= argc)
			radio(MEMORY);
		else {
			parse_freq(argv[arg]);
			radio(MEMORY, freq[4]); }
		break;
	case 18 :	radio(MEMORY_SET);		break;
	case 19 :	radio(MEMORY_VFO);		break;
	case 20 :	radio(MEMORY_CLR);		break;
	case 21 :	// Offset
		if(arg >= argc)
			printf("\nOffset required!\n");
		else {
			parse_freq(argv[arg]);
			radio(SET_OFFSET, freq[4], freq[3], freq[2]); }
		break;
	case 22 :	radio(SCAN_STOP);		break;
	case 23 :	radio(SCAN_START);		break;
	case 24 :	radio(SIMPLEX);			break;
	case 25 :	radio(DUPLEX_N);		break;
	case 26 :	radio(DUPLEX_P);		}

	/* All finished.. close comm port & exit */
	Cclose();
}
